/*
	BiDirectional linked list
	v1.2
	(c) 2002 by Nick Kochakian

    If you use any part of this library give me credit for writing it.
*/

#ifndef __BIDILIST_H
#define __BIDILIST_H

#ifdef WIN32
  #ifndef BIDI_LIST_CALLBACK
    #define BIDI_LIST_CALLBACK __cdecl
  #endif
#else
  #ifndef BIDI_LIST_CALLBACK
    #define BIDI_LIST_CALLBACK
  #endif
#endif

#ifdef __cplusplus
       #define BIDI_LIST_FUNCTION(returnType, functionName, arguments) extern returnType BIDI_LIST_CALLBACK functionName##arguments##;
#else
       #define BIDI_LIST_FUNCTION(returnType, functionName, arguments) returnType BIDI_LIST_CALLBACK functionName##arguments##;
#endif

#ifdef __cplusplus
extern "C" {
#endif

typedef struct BIDI_LIST_DATA
{
	void *data;

	struct BIDI_LIST_DATA *next;
	struct BIDI_LIST_DATA *last;
} BIDI_LIST_DATA;

typedef struct
{
	BIDI_LIST_DATA *listStart;
	BIDI_LIST_DATA *listEnd;
} BIDI_LIST;

BIDI_LIST_FUNCTION(BIDI_LIST *, BiDiListCreate, (void))
BIDI_LIST_FUNCTION(unsigned char, BiDiListAdd, (BIDI_LIST *list, void *data))
BIDI_LIST_FUNCTION(BIDI_LIST_DATA *, BiDiListRemove, (BIDI_LIST *list, BIDI_LIST_DATA *listData, void (*dataDestroyCallback)(void *data)))
BIDI_LIST_FUNCTION(void, BiDiListSwitchData, (BIDI_LIST *list, BIDI_LIST_DATA *a, BIDI_LIST_DATA *b))
BIDI_LIST_FUNCTION(void, BiDiListMoveToFront, (BIDI_LIST *list, BIDI_LIST_DATA *a))
BIDI_LIST_FUNCTION(void, BiDiListDestroy, (BIDI_LIST *list, void (*dataDestroyCallback)(void *data)))
BIDI_LIST_FUNCTION(BIDI_LIST *, BiDiListCopy, (BIDI_LIST *list))
BIDI_LIST_FUNCTION(void, BiDiListClear, (BIDI_LIST *list, void (*dataDestroyCallback)(void *data)))
BIDI_LIST_FUNCTION(int, BiDiListInsert, (BIDI_LIST *list, BIDI_LIST_DATA *position, void *data))
BIDI_LIST_FUNCTION(int, BiDiListInsertAfter, (BIDI_LIST *list, BIDI_LIST_DATA *position, void *data))

#ifdef __cplusplus
};
#endif

#endif
