/* 
 * This file is part of the dcload Dreamcast serial loader
 *
 * Copyright (C) 2001 Andrew Kieschnick <andrewk@napalm-x.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/********************************************************
 * Win32 port and transmission fixes by NickK 8/22/2002 *
 ********************************************************/

#include <conio.h>
#include "dc-tool.h"

struct
{
    unsigned long address;
    unsigned long size;
    unsigned char *filename;
    unsigned char *path;
    unsigned char console;
    unsigned char dumbTerm;
    unsigned char quiet;
    unsigned long speed;
    char          *deviceName;
    unsigned char cdfsRedir;
    unsigned char *isoFile;
    unsigned char speedHack;
    HANDLE        hCommPort;
} cfg;

#ifndef HAVE_GETOPT
/* The following code for getopt is from the libc-source of FreeBSD,
 * it might be changed a little bit.
 * Florian Schulze (florian.proff.schulze@gmx.net)
 */

/*
 * Copyright (c) 1987, 1993, 1994
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

int     opterr = 1,             /* if error message should be printed */
        optind = 1,             /* index into parent argv vector */
        optopt,                 /* character checked for validity */
        optreset;               /* reset getopt */
char    *optarg;                /* argument associated with option */

#define BADCH   (int)'?'
#define BADARG  (int)':'
#define EMSG    ""

char *__progname="prboom_server";

/*
 * getopt --
 *      Parse argc/argv argument vector.
 */
int
getopt(nargc, nargv, ostr)
int nargc;
char * const *nargv;
const char *ostr;
{
    extern char *__progname;
    static char *place = EMSG;              /* option letter processing */
    char *oli;                              /* option letter list index */
    int ret;
    
    if (optreset || !*place) {              /* update scanning pointer */
	optreset = 0;
	if (optind >= nargc || *(place = nargv[optind]) != '-') {
	    place = EMSG;
	    return (-1);
	}
	if (place[1] && *++place == '-') {      /* found "--" */
	    ++optind;
	    place = EMSG;
	    return (-1);
	}
    }                                       /* option letter okay? */
    if ((optopt = (int)*place++) == (int)':' ||
	!(oli = strchr(ostr, optopt))) {
	/*
	 * if the user didn't specify '-' as an option,
	 * assume it means -1.
	 */
	if (optopt == (int)'-')
	    return (-1);
	if (!*place)
	    ++optind;
	if (opterr && *ostr != ':')
	    (void)fprintf(stderr,
			  "%s: illegal option -- %c\n", __progname, optopt);
	return (BADCH);
    }
    if (*++oli != ':') {                    /* don't need argument */
	optarg = NULL;
	if (!*place)
	    ++optind;
    }
    else {                                  /* need an argument */
	if (*place)                     /* no white space */
	    optarg = place;
	else if (nargc <= ++optind) {   /* no arg */
	    place = EMSG;
	    if (*ostr == ':')
		ret = BADARG;
	    else
		ret = BADCH;
	    if (opterr)
		(void)fprintf(stderr,
			      "%s: option requires an argument -- %c\n",
			      __progname, optopt);
	    return (ret);
	}
	else                            /* white space */
	    optarg = nargv[optind];
	place = EMSG;
	++optind;
    }
    return (optopt);                        /* dump back option letter */
}
#else
#include <unistd.h>
#endif

void usage(void)
{
    printf("\ndc-tool %s by <andrewk@napalm-x.com>\n", VERSION);
    printf("Modified by NickK\n\n");
    printf("-x <filename> Upload and execute <filename>\n");
    printf("-u <filename> Upload <filename>\n");
    printf("-d <filename> Download to <filename>\n");
    printf("-a <address>  Set address to <address> (default: 0x8C010000)\n");
    printf("-s <size>     Set size to <size>\n");
    printf("-t <device>   Use <device> to communicate with dc (default: %s)\n", SERIAL_DEVICE);
    printf("-b <baudrate> Use <baudrate> (default: %d)\n", BAUD_RATE);
    printf("-e            Try alternate 115200 (must also use -b 115200)\n");
    printf("-n            Do not attach console and fileserver\n");
    printf("-p            Use dumb terminal rather than console/fileserver\n");
    printf("-q            Do not clear screen before download\n");
    printf("-c <path>     Chroot to <path> (must be super-user)\n");
    printf("-i <isofile>  Enable cdfs redirection using iso image <isofile>\n");
    printf("-h            Usage information (you\'re looking at it)\n\n");
}

unsigned int upload(unsigned char *filename, unsigned int address)
{
    int           size;
    unsigned char *inbuf;
    unsigned long startTime, endTime;
    unsigned char c;
    float         stime, etime;
    bfd           *somebfd;
    FILE          *f;
    asection      *section;

    lzo_init();

    somebfd = bfd_openr(filename, 0);

    if (somebfd) /* try bfd first */
    {
	   if (bfd_check_format(somebfd, bfd_object))
       {
	      printf("File format is %s, ", somebfd->xvec->name);
	      address = somebfd->start_address;
	      size    = 0;
	      printf("start address is 0x%x\n", address);

          startTime = GetTickCount();

	      for (section = somebfd->sections; section != NULL; section = section->next)
          {
		      if ((section->flags & SEC_HAS_CONTENTS) &&
                  (section->flags & SEC_LOAD))
              {
		         printf("Section %s, ",section->name);
		         printf("lma 0x%x, ",section->lma);
		         printf("size %d\n",section->_raw_size);

		         if (section->_raw_size)
                 {
			        size += section->_raw_size;
			        inbuf = malloc(section->_raw_size);
			        bfd_get_section_contents(somebfd, section, inbuf,
                                             0, section->_raw_size);

                    c = 'B';
                    if (!serialWrite(cfg.hCommPort, &c, 1))
                       printf("Write error\n");

                    if (!serialRead(cfg.hCommPort, &c, 1))
                       printf("Read error\n");

                    if (!sendUInt(cfg.hCommPort, section->lma))
                       printf("Write error\n");

                    if (!sendUInt(cfg.hCommPort, section->_raw_size))
                       printf("Write error\n");

                    if (!sendData(cfg.hCommPort, inbuf,
                                  section->_raw_size, 1))
                       printf("Write error\n");

			        free(inbuf);
                 }
              }
          }
       } else
       {
          /* if that fails, send raw bin */
		    
	      printf("File format is raw binary, start address is 0x%x\n", address);

          f = fopen(filename, "rb");

          if (!f)
          {
             printf("Could not open '%s'!\n", filename);
             bfd_close(somebfd);
             exit(-1);
          }

          fseek(f, 0, SEEK_END);
          size = ftell(f);
          fseek(f, 0, SEEK_SET);

	      inbuf = malloc(size);

          fread(inbuf, 1, size, f);

          fclose(f);

          startTime = GetTickCount();
	    
	      c = 'B';

          if (!serialWrite(cfg.hCommPort, &c, 1))
             printf("Write error\n");

          if (!serialRead(cfg.hCommPort, &c, 1))
             printf("Read error\n");

          if (!sendUInt(cfg.hCommPort, address))
             printf("Write error\n");

          if (!sendUInt(cfg.hCommPort, size))
             printf("Write error\n");

          if (!sendData(cfg.hCommPort, inbuf, size, 1))
             printf("Write error\n");
       }

       endTime = GetTickCount();

       stime = (float)startTime / 1000000000.0f;
       etime = (float)endTime / 1000000000.0f;

       printf("effective: %f bytes / sec\n", (float)size / (etime - stime));
       fflush(stdout);
    }

    bfd_close(somebfd);

    return address;
}

void download(unsigned char *filename, unsigned int address,
              unsigned int size, unsigned int quiet)
{
    unsigned char c;
    unsigned int  wrkmem = 0x8cff0000;
    unsigned char *data;
    unsigned long startTime, endTime;
    float         stime, etime;
    FILE          *f;

    f = fopen(filename, "wb");
    if (!f)
    {
       printf("Could not open the file '%s' for writing!\n", filename);
       exit(-1);
    }

    data = malloc(size);
    c    = (!quiet) ? 'F' : 'G';

    if (!serialWrite(cfg.hCommPort, &c, 1))
       printf("Write error\n");

    if (!serialRead(cfg.hCommPort, &c, 1))
       printf("Read error\n");

    if (!sendUInt(cfg.hCommPort, address))
       printf("Write error\n");

    if (!sendUInt(cfg.hCommPort, size))
       printf("Write error\n");

    if (!sendUInt(cfg.hCommPort, wrkmem))
       printf("Write error\n");

    startTime = GetTickCount();

    if (!recvData(cfg.hCommPort, data, size, 1))
       printf("Read error\n");

    endTime = GetTickCount();    
    
    printf("Received %d bytes\n", size);

    stime = (float)startTime / 1000000000.0f;
    etime = (float)endTime / 1000000000.0f;

    printf("effective: %f bytes / sec\n", (float)size / (etime - stime));
    fflush(stdout);

    fwrite(data, 1, size, f);
    fclose(f);
}

void execute(unsigned int address, unsigned int console)
{
    unsigned char c;

    printf("Sending execute command (0x%x, console=%d)...",address,console);

    c = 'A';

    if (!serialWrite(cfg.hCommPort, &c, 1))
       printf("Write error\n");

    if (!serialRead(cfg.hCommPort, &c, 1))
       printf("Read error\n");

    if (!sendUInt(cfg.hCommPort, address))
       printf("Write error\n");

    if (!sendUInt(cfg.hCommPort, console))
       printf("Write error\n");

    printf("executing\n");
}

void do_console(unsigned char *path, unsigned char *isofile)
{
    unsigned char command;
    unsigned long commEvent = 0;
    FILE          *f        = NULL;

    if (cfg.isoFile)
    {
       f = fopen(cfg.isoFile, "rb");
       if (!f)
       {
          printf("Could not open the file '%s'!\n", cfg.isoFile);
          return;
       }
    }

    while (1)
    {
          if (serialRead(cfg.hCommPort, &command, 1) > 0)
          {
             fflush(stdout);

             switch (command)
             {
                    case 0:
                         FlushFileBuffers(cfg.hCommPort);
                         if (f != NULL)
                            fclose(f);
                    return;

                    case 1:
                         dcFStat(cfg.hCommPort);
                    break;

                    case 2:
                         dcWrite(cfg.hCommPort);
                    break;

                    case 3:
                         dcRead(cfg.hCommPort);
                    break;

                    case 4:
                         dcOpen(cfg.hCommPort);
                    break;

                    case 5:
                         dcClose(cfg.hCommPort);
                    break;

                    case 6:
                         dcCreate(cfg.hCommPort);
                    break;

                    case 7:
                         dcLink(cfg.hCommPort);
                    break;

                    case 8:
                         dcUnlink(cfg.hCommPort);
                    break;

                    case 9:
                         dcChdir(cfg.hCommPort);
                    break;

                    case 10:
                         dcChmod(cfg.hCommPort);
                    break;

                    case 11:
                         dcLSeek(cfg.hCommPort);
                    break;

                    case 12:
                         dcTime(cfg.hCommPort);
                    break;

                    case 13:
                         dcStat(cfg.hCommPort);
                    break;

                    case 14:
                         dcUTime(cfg.hCommPort);
                    break;

                    case 15:
                         printf("Command 15 - ERROR\n");
                    break;

                    case 16:
                         dcOpenDir(cfg.hCommPort);
                    break;

                    case 17:
                         dcCloseDir(cfg.hCommPort);
                    break;

                    case 18:
                         dcReadDir(cfg.hCommPort);
                    break;

                    case 19:
                         dcCDFSRedirReadSectors(cfg.hCommPort, f);
                    break;

                    default:
                        printf("Bad command: %d\n");
                        printf("dc-tool will now terminate\n");
                        printf("Press any key to continue\n");

                        FlushFileBuffers(cfg.hCommPort);

                        if (f != NULL)
                           fclose(f);

                        while (!kbhit());
                    return;
             }
          }
    }
}

/* dumb terminal mode
 * for programs that don't use dcload I/O functions
 * FIXME: should allow setting a different baud rate from what dcload uses
 * FIXME: should allow input
 */

void do_dumbterm(void)
{
    unsigned char c;

    printf("\nDumb terminal mode isn't implemented, so there. Bye.\n\n"); //you get this half-assed one.\n\n");

/*    while (1) {
	blread(&c, 1);
	printf("%c", c);
    }*/
}

void exitCallback(void)
{
     if (cfg.filename != NULL)
     {
        free(cfg.filename);
        cfg.filename = NULL;
     }

     if (cfg.path != NULL)
     {
        free(cfg.path);
        cfg.path = NULL;
     }

     if (cfg.deviceName != NULL)
     {
        free(cfg.deviceName);
        cfg.deviceName = NULL;
     }

     if (cfg.isoFile != NULL)
     {
        free(cfg.isoFile);
        cfg.isoFile = NULL;
     }

     if (cfg.hCommPort != NULL)
     {
        closeSerialPort(cfg.hCommPort);
        cfg.hCommPort = NULL;
     }

     fcloseall();
     dcLoadSysCallsShutdown();
}

int main(int argc, char *argv[])
{
    unsigned char command = 0;
    int           someopt;
    struct stat   statBuf;
    unsigned char c;

    if (argc < 2)
    {
	   usage();
       return 1;
    }

    /* Set the default values in the configuration structure */
    cfg.address    = 0x8C010000;
    cfg.size       = 0;
    cfg.filename   = NULL;
    cfg.path       = NULL;
    cfg.console    = 1;
    cfg.dumbTerm   = 0;
    cfg.quiet      = 0;
    cfg.speed      = BAUD_RATE;
    cfg.deviceName = malloc(strlen(SERIAL_DEVICE) + 1);
    cfg.cdfsRedir  = 0;
    cfg.isoFile    = NULL;
    cfg.speedHack  = 0;
    cfg.hCommPort  = NULL;

    sprintf(cfg.deviceName, "%s", SERIAL_DEVICE);

    atexit(exitCallback);

    dcLoadSysCallsInit();

    someopt = getopt(argc, argv, "x:u:d:a:s:t:b:c:i:npqhe");
    while (someopt > 0)
    {
          switch (someopt)
          {
                 case 'x':
                      if (command)
                      {
                         printf("You can only specify one of -x, -u, and -d\n");
                         return 1;
	                  }

                      if (cfg.filename != NULL)
                         free(cfg.filename);

                      command      = 'x';
                      cfg.filename = malloc(strlen(optarg) + 1);
                      strcpy(cfg.filename, optarg);
                 break;

                 case 'u':
	                  if (command)
                      {
                         printf("You can only specify one of -x, -u, and -d\n");
                         return 1;
                      }

                      if (cfg.filename != NULL)
                         free(cfg.filename);

                      command      = 'u';
                      cfg.filename = malloc(strlen(optarg) + 1);
                      strcpy(cfg.filename, optarg);
                 break;

                 case 'd':
                      if (command)
                      {
                         printf("You can only specify one of -x, -u, and -d\n");
                         return 1;
                      }

                      if (cfg.filename != NULL)
                         free(cfg.filename);

                      command      = 'd';
                      cfg.filename = malloc(strlen(optarg) + 1);
                      strcpy(cfg.filename, optarg);
                 break;

                 case 'c':
                      if (cfg.path != NULL)
                         free(cfg.path);

                      cfg.path = malloc(strlen(optarg) + 1);
                      strcpy(cfg.path, optarg);
                 break;

                 case 'i':
                      if (cfg.isoFile != NULL)
                         free(cfg.isoFile);

                      cfg.cdfsRedir = 1;
                      cfg.isoFile   = malloc(strlen(optarg) + 1);
                      strcpy(cfg.isoFile, optarg);
                 break;

                 case 'a':
                      cfg.address = strtoul(optarg, NULL, 0);
                 break;

                 case 's':
                      cfg.size = strtoul(optarg, NULL, 0);
                 break;

                 case 't':
                      if (cfg.deviceName != NULL)
                         free(cfg.deviceName);

                      cfg.deviceName = malloc(strlen(optarg) + 1);
                      strcpy(cfg.deviceName, optarg);
                 break;

                 case 'b':
                      cfg.speed = strtoul(optarg, NULL, 0);
                 break;

                 case 'n':
                      cfg.console = 0;
                 break;

                 case 'p':
                      cfg.console  = 0;
                      cfg.dumbTerm = 1;
                 break;

                 case 'q':
                      cfg.quiet = 1;
                 break;

                 case 'h':
                      usage();
                 return 0;

                 case 'e':
                      cfg.speedHack = 1;
                 break;
          }

          someopt = getopt(argc, argv, "x:u:d:a:s:t:b:c:i:npqhe");
    }

    if (command=='x' || command=='u')
    {
       if (stat(cfg.filename, &statBuf))
       {
          perror(cfg.filename);
          return 1;
       }
    }

    if (cfg.console)
       printf("Console enabled\n");

    if (cfg.dumbTerm)
       printf("Dumb terminal enabled\n");

    if (cfg.quiet)
       printf("Quiet download\n");

    if (cfg.path)
       printf("Chroot enabled\n");

    if (cfg.cdfsRedir)
       printf("Cdfs redirection enabled\n");

    if (cfg.speedHack)
       printf("Alternate 115200 enabled\n");

    cfg.hCommPort = openSerialPort(cfg.deviceName, BAUD_RATE);
    if (!cfg.hCommPort)
    {
       printf("Error opening serial device\n");
       return 1;
    }

    if (cfg.speed != BAUD_RATE)
       changeSerialPortSpeed(cfg.hCommPort, cfg.speed, cfg.speedHack);

    switch (command)
    {
           case 'x':
                if (cfg.cdfsRedir)
                {
                   c = 'H';
                   if (!serialWrite(cfg.hCommPort, &c, 1))
                      printf("Write error\n");

                   if (!serialRead(cfg.hCommPort, &c, 1))
                      printf("Read error\n");
                }

                printf("Upload <%s>\n", cfg.filename);
                cfg.address = upload(cfg.filename, cfg.address);
                printf("Executing at <0x%x>\n", cfg.address);
                execute(cfg.address, cfg.console);

                if (cfg.console)
                   do_console(cfg.path, cfg.isoFile);
                   else if (cfg.dumbTerm)
                           do_dumbterm();
           break;

           case 'u':
                printf("Upload <%s> at <0x%x>\n", cfg.filename,
                                                  cfg.address);
                upload(cfg.filename, cfg.address);
                changeSerialPortSpeed(cfg.hCommPort, BAUD_RATE,
                                      cfg.speedHack);
           break;

           case 'd':
                if (!cfg.size)
                {
                   printf("You must specify a size (-s <size>) with download (-d <filename>)\n");
                   return 1;
                }

                printf("Download %d bytes at <0x%x> to <%s>\n", cfg.size,
                       cfg.address, cfg.filename);

                download(cfg.filename, cfg.address, cfg.size,
                         cfg.quiet);

                changeSerialPortSpeed(cfg.hCommPort, BAUD_RATE,
                                      cfg.speedHack);
           break;

           default:
             usage();
           return 1;
    }

    return 0;
}

/* This took some time to write so now I'm not going to delete it
   :P */

/*#ifdef _WINDOWS
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                   LPSTR lpCmdLine, int nShowCmd)
{
    int  argc   = 0;
    int  startOffset;
    char **argv = NULL;
    int  i;
    int  l;
    int  cmdLineLength = strlen(lpCmdLine);

    /* Get argc count */
/*    for (i=0; i<=strlen(lpCmdLine); i++)
    {
        if (lpCmdLine[i]==' ' || lpCmdLine[i]==0)
           argc++;
    }

    /* Allocate argv */
/*    argv        = (char **)malloc((argc + 1) * sizeof(unsigned long));
    startOffset = 0;
    l           = 1;

    argv[0] = &lpCmdLine[0];

    for (i=0; i<=cmdLineLength; i++)
    {
        if (lpCmdLine[i]==' ' || lpCmdLine[i]==0)
        {
           argv[l]      = &lpCmdLine[startOffset];
           lpCmdLine[i] = 0;
           l++;
           startOffset = i + 1;
        }
    }

    main(argc, argv);

    /* Deallocate argv */
/*    free(argv);

    return 0;
}
#endif*/