/* 
 * This file is part of the dcload Dreamcast serial loader
 *
 * Copyright (C) 2001 Andrew Kieschnick <andrewk@napalm-x.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/********************************************************
 * Win32 port and transmission fixes by NickK 8/22/2002 *
 ********************************************************/

#include "dc-tool.h"

#define HEAP_ALLOC(var,size) \
        long __LZO_MMODEL var [ ((size) + (sizeof(long) - 1)) / sizeof(long) ]

static HEAP_ALLOC(wrkmem, LZO1X_1_MEM_COMPRESS);

HANDLE openSerialPort(char *portName, unsigned long baudRate)
{
    HANDLE hCommPort = NULL;
    DCB    portConfig;
    COMMPROP commSettings;
    unsigned long rxBufferSize = 8192;
    unsigned long txBufferSize = 8192;

    hCommPort = CreateFile(portName,
                           GENERIC_READ | GENERIC_WRITE,
                           0,     /* Open with exclusive access */
                           NULL,  /* No security attributes */
                           OPEN_EXISTING,
                           0,     /* Not using overlapped I/O */
                           NULL); /* Must be NULL for comm devices */

    if (hCommPort==INVALID_HANDLE_VALUE)
       return NULL;

    portConfig.DCBlength = sizeof(DCB);

    GetCommProperties(hCommPort, &commSettings);
    if (commSettings.dwMaxRxQueue > 0 &&
        rxBufferSize > commSettings.dwMaxRxQueue)
       rxBufferSize = commSettings.dwMaxRxQueue;

    if (commSettings.dwMaxTxQueue > 0 &&
        txBufferSize > commSettings.dwMaxTxQueue)
       txBufferSize = commSettings.dwMaxTxQueue;

    if (rxBufferSize < commSettings.dwCurrentRxQueue)
       rxBufferSize = commSettings.dwCurrentRxQueue;

    if (txBufferSize < commSettings.dwCurrentTxQueue)
       txBufferSize = commSettings.dwCurrentTxQueue;

    SetupComm(hCommPort, rxBufferSize, txBufferSize);

    if (!GetCommState(hCommPort, &portConfig))
    {
       printf("Error: Could not get the state of %s!\n", portName);

       CloseHandle(hCommPort);
       return NULL;
    }

    portConfig.BaudRate          = baudRate;
    portConfig.ByteSize          = DATA_BITS;
    portConfig.Parity            = PARITY_SET;
    portConfig.StopBits          = STOP_BITS;
    /*portConfig.fDsrSensitivity   = FALSE;
    portConfig.fOutxCtsFlow      = TRUE;
    portConfig.fOutxDsrFlow      = FALSE;
    portConfig.fDtrControl       = DTR_CONTROL_DISABLE;
    portConfig.fDsrSensitivity   = FALSE;
    portConfig.fTXContinueOnXoff = FALSE;
    portConfig.fOutX             = FALSE;
    portConfig.fInX              = FALSE;
    portConfig.fNull             = FALSE;
    portConfig.fRtsControl       = RTS_CONTROL_TOGGLE;
    portConfig.fAbortOnError     = FALSE;*/

    if (!SetCommState(hCommPort, &portConfig))
    {
       printf("Error: Could not set the state of %s!\n", portName);

       CloseHandle(hCommPort);
       return NULL;
    }

    SetCommMask(hCommPort, EV_CTS | EV_RXCHAR | EV_RXFLAG |
                           EV_TXEMPTY | EV_BREAK | EV_ERR);

    /* Clear the communication port's buffers */
    PurgeComm(hCommPort, PURGE_TXCLEAR | PURGE_RXCLEAR);

    return hCommPort;
}

void closeSerialPort(HANDLE hCommPort)
{
     if (hCommPort != NULL)
        CloseHandle(hCommPort);
}

int changeSerialPortSpeed(HANDLE hCommPort, unsigned long baudRate,
                          unsigned char speedHack)
{
    DCB           portConfig;
    unsigned char c;

    if (!hCommPort)
       return 0;

    if (!GetCommState(hCommPort, &portConfig))
       return 0;

    c = 'S';
    if (!serialWrite(hCommPort, &c, 1))
       return 0;

    if (speedHack && baudRate==115200)
       sendUInt(hCommPort, 111600); /* get dcload to pick N=13 rather than N=12 */
       else
       sendUInt(hCommPort, baudRate);

    portConfig.BaudRate = baudRate;

    if (!SetCommState(hCommPort, &portConfig))
       return 0;

    return 1;
}

/*int serialReadDirect(HANDLE hCommPort, void *buffer, int count)
{
    unsigned long actualBytesRead = 0;

    if (!ReadFile(hCommPort, buffer, count, &actualBytesRead, NULL))
       return 0;

    return actualBytesRead;
}*/

/*
#define EV_RXCHAR           0x0001  // Any Character received
#define EV_RXFLAG           0x0002  // Received certain character
#define EV_TXEMPTY          0x0004  // Transmitt Queue Empty
#define EV_CTS              0x0008  // CTS changed state
#define EV_DSR              0x0010  // DSR changed state
#define EV_RLSD             0x0020  // RLSD changed state
#define EV_BREAK            0x0040  // BREAK received
#define EV_ERR              0x0080  // Line status error occurred
#define EV_RING             0x0100  // Ring signal detected
#define EV_PERR             0x0200  // Printer error occured
#define EV_RX80FULL         0x0400  // Receive buffer is 80 percent full
#define EV_EVENT1           0x0800  // Provider specific event 1
#define EV_EVENT2           0x1000  // Provider specific event 2

#define MS_CTS_ON           ((DWORD)0x0010)
#define MS_DSR_ON           ((DWORD)0x0020)
#define MS_RING_ON          ((DWORD)0x0040)
#define MS_RLSD_ON          ((DWORD)0x0080)
*/

int serialRead(HANDLE hCommPort, void *buffer, int count)
{
    unsigned long actualBytesRead = 0;
    unsigned long totalBytesRead  = 0;
    unsigned char *data           = (unsigned char *)buffer;
    int           result          = 0;
    unsigned long commEvent       = 0;
    unsigned long commErrors      = 0;
    COMSTAT       commStat;
    COMMPROP      commSettings;

    GetCommProperties(hCommPort, &commSettings);

    result = 0;

    while (count > 0)
    {
          actualBytesRead = 0;
          result          = 0;

          /* Check for errors */
          GetCommMask(hCommPort, &commEvent);

          if (commEvent & EV_BREAK)
             ClearCommBreak(hCommPort);
          if (commEvent & EV_ERR)
             ClearCommError(hCommPort, &commErrors, &commStat);

          if (commSettings.dwCurrentRxQueue > 0)
          {
             if ((unsigned long)count < commSettings.dwCurrentRxQueue)
             {
                /* Wait for the incoming data to completely arrive before
                   reading it */

                do
                {
                    ClearCommError(hCommPort, &commErrors, &commStat);
                } while (commStat.cbInQue < (unsigned long)count);
             } else
             {
                /* Since the data that we're reading is larger than the
                   incoming transmission buffer, wait for the buffer to
                   completely fill before reading anything */

                do
                {
                    ClearCommError(hCommPort, &commErrors, &commStat);
                } while (commStat.cbInQue < commSettings.dwCurrentRxQueue);
             }
          }

          result = ReadFile(hCommPort, data, count, &actualBytesRead,
                            NULL);

          /*if (!result && count > 1)
             result = ReadFile(hCommPort, data, 1, &actualBytesRead,
                               NULL);*/

          if (result)
          {
             data += actualBytesRead;
             count -= actualBytesRead;
             totalBytesRead += actualBytesRead;
          }
    }

    return totalBytesRead;
}

int serialWrite(HANDLE hCommPort, void *buffer, int count)
{
    unsigned long actualBytesWritten = 0;
    unsigned long totalBytesWritten  = 0;
    unsigned char *data              = (unsigned char *)buffer;
    unsigned long commEvent          = 0;
    unsigned long commErrors         = 0;
    COMSTAT       commStat;

    while (count > 0)
    {
          /* Check for errors */
          GetCommMask(hCommPort, &commEvent);

          if (commEvent & EV_BREAK)
             ClearCommBreak(hCommPort);
          if (commEvent & EV_ERR)
             ClearCommError(hCommPort, &commErrors, &commStat);

          actualBytesWritten = 0;
          if (WriteFile(hCommPort, data, count, &actualBytesWritten,
                        NULL) && actualBytesWritten > 0)
          {
             data += actualBytesWritten;
             count -= actualBytesWritten;
             totalBytesWritten += actualBytesWritten;
          }
    }

    return totalBytesWritten;
}

int sendUInt(HANDLE hCommPort, unsigned int value)
{
    unsigned int dcValue = 0;

    if (!serialWrite(hCommPort, &value, 4))
       return 0;

    if (!serialRead(hCommPort, &dcValue, 4))
       return 0;

    if (dcValue != value)
       return 0;

    return 1;
}

int recvUInt(HANDLE hCommPort, unsigned int *value)
{
    unsigned char *data = (unsigned char *)value;
    int           count = 4;

    while (count > 0)
    {
          if (serialRead(hCommPort, data, 1))
          {
             data++;
             count--;
          }
    }

    return 1;
}

/* receive total bytes from dc and store in data */
int recvData(HANDLE hCommPort, void *data, unsigned int total,
             unsigned int verbose)
{
     unsigned char type, sum, ok;
     unsigned int  size, newsize;
     unsigned char *tmp;

     if (verbose)
     { 
	    printf("recv_data: ");
	    fflush(stdout);
     }

     while (total > 0)
     {
           if (!serialRead(hCommPort, &type, 1))
              return 0;

           if (!recvUInt(hCommPort, &size))
              return 0;

           switch (type)
           {
                  case 'U': /* Uncompressed */
                       if (verbose)
                       {
                          printf("U");
                          fflush(stdout);
                       }

                       if (!serialRead(hCommPort, data, size))
                          return 0;

                       if (!serialRead(hCommPort, &sum, 1))
                          return 0;

                       ok = 'G';

                       if (!serialWrite(hCommPort, &ok, 1))
                       {
                          printf("Write error #1\n");
                          return 0;
                       }

                       total -= size;
                       (unsigned char *)data += size;
                  break;

                  case 'C': /* Compressed */
                       if (verbose)
                       {
                          printf("C");
                          fflush(stdout);
                       }

                       tmp = malloc(size);
                       if (!tmp)
                       {
                          printf("Out of memory for a compressed data buffer!\n");
                          return 0;
                       }

                       if (!serialRead(hCommPort, tmp, size))
                       {
                          free(tmp);
                          return 0;
                       }

                       if (!serialRead(hCommPort, &sum, 1))
                       {
                          free(tmp);
                          return 0;
                       }

                       if (lzo1x_decompress(tmp, size, data, &newsize,
                                            0)==LZO_E_OK)
                       {
                          ok = 'G';
                          if (!serialWrite(hCommPort, &ok, 1))
                          {
                             free(tmp);
                             printf("Write error #2\n");
                             return 0;
                          }

                          total -= newsize;
                          (unsigned char *)data += newsize;
                       } else
                       {
                          ok = 'B';
                          if (!serialWrite(hCommPort, &ok, 1))
                          {
                             free(tmp);
                             printf("Write error #3\n");
                             return 0;
                          }

                          printf("\nrecvData: Decompression failed!\n");
                       }

                       free(tmp);
                  break;
           }
     }

     if (verbose)
     {
        printf("\n");
        fflush(stdout);
     }

     return 1;
}

/* send size bytes to dc from addr */
int sendData(HANDLE hCommPort, unsigned char *addr, unsigned int size,
             unsigned int verbose)
{
    unsigned int i;
    unsigned char *location = (unsigned char *)addr;
    unsigned char sum = 0;
    unsigned char data;
    unsigned int csize;
    unsigned int sendsize;
    unsigned char c;
    unsigned char *buffer;

    buffer = malloc(DCLOAD_BUFFER + DCLOAD_BUFFER / 64 + 16 + 3);
    if (!buffer)
    {
       printf("Error in sendData: Out of memory for a data buffer!\n");
       return 0;
    }

    if (verbose)
    {
       printf("sendData: ");
       fflush(stdout);
    }

    while (size > 0)
    {
          if (size > 8192)
             sendsize = 8192;
             else
             sendsize = size;

          lzo1x_1_compress((unsigned char *)addr, sendsize, buffer,
                           &csize, wrkmem);

          if (csize < sendsize)
          {
             /* Send compressed data */

             if (verbose)
             {
                printf("C");
                fflush(stdout);
             }

             c = 'C';

             if (!serialWrite(hCommPort, &c, 1))
             {
                free(buffer);
                return 0;
             }

             if (!sendUInt(hCommPort, csize))
             {
                free(buffer);
                return 0;
             }

             data = 'B';
             while (data != 'G')
             {
                   location = buffer;
                   if (!serialWrite(hCommPort, location, csize))
                   {
                      free(buffer);
                      return 0;
                   }

                   sum = 0;
                   for (i=0; i<csize; i++)
                   {
                       data = *location++;
                       sum ^= data;
                   }

                   if (!serialWrite(hCommPort, &sum, 1))
                   {
                      free(buffer);
                      return 0;
                   }

                   if (!serialRead(hCommPort, &data, 1))
                   {
                      free(buffer);
                      printf("Read error #1 in sendData!\n");
                      return 0;
                   }
             }
          } else
          {
             /* Send the data uncompressed since compression, in this
                case, has increased the size of the uncompressed
                data */

             if (verbose)
             {
                printf("U");
                fflush(stdout);
             }

             c = 'U';
             if (!serialWrite(hCommPort, &c, 1))
             {
                free(buffer);
                return 0;
             }

             if (!sendUInt(hCommPort, sendsize))
             {
                free(buffer);
                return 0;
             }

             if (!serialWrite(hCommPort, addr, sendsize))
             {
                free(buffer);
                return 0;
             }

             sum = 0;
             for (i=0; i<sendsize; i++)
                 sum ^= ((unsigned char *)addr)[i];

             if (!serialWrite(hCommPort, &sum, 1))
             {
                free(buffer);
                return 0;
             }

             if (!serialRead(hCommPort, &data, 1))
             {
                free(buffer);
                printf("Read error #2 in sendData!\n");
                return 0;
             }
          }

          size -= sendsize;
          addr += sendsize;
    }

    if (verbose)
    {
       printf("\n");
       fflush(stdout);
    }

    free(buffer);

    return 1;
}

void outputError(void)
{
    char *lpMsgBuf;
    
    FormatMessage(
	FORMAT_MESSAGE_ALLOCATE_BUFFER |
	FORMAT_MESSAGE_FROM_SYSTEM |
	FORMAT_MESSAGE_IGNORE_INSERTS,
	NULL,
	GetLastError(),
	MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
	(LPTSTR) &lpMsgBuf, 0, NULL // Process any inserts in lpMsgBuf.
	);
    printf("%s\n",(char *)lpMsgBuf);
    LocalFree( (LPVOID) lpMsgBuf );
}
